@@ -4,7 +4,7 @@ import re
 
 
 # Find similar string from a list of strings
-def find_similar_str(x, lookup_list, processor='fuzzywuzzy', substitution_cost=100):
+def find_similar_str(x, lookup_list, processor='fuzzywuzzy', **kwargs):
     """
     :param x: [str]
     :param lookup_list: [iterable]
@@ -16,12 +16,12 @@ def find_similar_str(x, lookup_list, processor='fuzzywuzzy', substitution_cost=1
 
     if processor == 'fuzzywuzzy':
         import fuzzywuzzy.fuzz
-        l_distances = [fuzzywuzzy.fuzz.token_set_ratio(x, a) for a in lookup_list]
+        l_distances = [fuzzywuzzy.fuzz.token_set_ratio(x, a, **kwargs) for a in lookup_list]
         the_one = lookup_list[l_distances.index(max(l_distances))] if l_distances else None
 
     elif processor == 'nltk':
         import nltk.metrics
-        l_distances = [nltk.edit_distance(x, a, substitution_cost=substitution_cost) for a in lookup_list]
+        l_distances = [nltk.edit_distance(x, a, **kwargs) for a in lookup_list]
         the_one = lookup_list[l_distances.index(min(l_distances))] if l_distances else None
 
     else:
@@ -39,7 +39,7 @@ def find_matched_str(x, lookup_list):
     """
     # assert isinstance(x, str), "'x' must be a string."
     # assert isinstance(lookup, list) or isinstance(lookup, tuple), "'lookup' must be a list/tuple"
-    if x is '' or x is None:
+    if x == '' or x is None:
         return None
     else:
         for y in lookup_list:
